@@ -3,7 +3,7 @@
  *
  * The connection object - this file is part a of the C extension module.
  *
- * Copyright (c) 2020 by the PyGreSQL Development Team
+ * Copyright (c) 2022 by the PyGreSQL Development Team
  *
  * Please see the LICENSE.TXT file for specific restrictions.
  */
@@ -511,7 +511,7 @@ conn_describe_prepared(connObject *self, PyObject *args)
     /* reads args */
     if (!PyArg_ParseTuple(args, "s#", &name, &name_length)) {
         PyErr_SetString(PyExc_TypeError,
-                        "Method prepare() takes a string argument");
+                        "Method describe_prepared() takes a string argument");
         return NULL;
     }
 
@@ -641,7 +641,9 @@ conn_set_non_blocking(connObject *self, PyObject *args)
     }
 
     if (!PyArg_ParseTuple(args, "i", &non_blocking)) {
-        PyErr_SetString(PyExc_TypeError, "setnonblocking(tf), with boolean.");
+        PyErr_SetString(
+            PyExc_TypeError,
+            "set_non_blocking() expects a boolean value as argument");
         return NULL;
     }
 
@@ -658,7 +660,7 @@ static char conn_is_non_blocking__doc__[] =
 "is_non_blocking() -- report the blocking status of the connection";
 
 static PyObject *
-conn_is_non_blocking(connObject *self, PyObject *args)
+conn_is_non_blocking(connObject *self, PyObject *noargs)
 {
     int rc;
 
@@ -667,41 +669,32 @@ conn_is_non_blocking(connObject *self, PyObject *args)
         return NULL;
     }
 
-    if (!PyArg_ParseTuple(args, "")) {
-        PyErr_SetString(PyExc_TypeError,
-            "method is_non_blocking() takes no parameters");
-        return NULL;
-    }
-
     rc = PQisnonblocking(self->cnx);
     if (rc < 0) {
         PyErr_SetString(PyExc_IOError, PQerrorMessage(self->cnx));
         return NULL;
     }
 
-    return PyBool_FromLong(rc);
+    return PyBool_FromLong((long)rc);
 }
 #endif /* DIRECT_ACCESS */
 
 
 /* Insert table */
 static char conn_inserttable__doc__[] =
-"inserttable(table, data, [columns]) -- insert list into table\n\n"
-"The fields in the list must be in the same order as in the table\n"
-"or in the list of columns if one is specified.\n";
+"inserttable(table, data, [columns]) -- insert iterable into table\n\n"
+"The fields in the iterable must be in the same order as in the table\n"
+"or in the list or tuple of columns if one is specified.\n";
 
 static PyObject *
 conn_inserttable(connObject *self, PyObject *args)
 {
     PGresult *result;
-    char *table, *buffer, *bufpt;
+    char *table, *buffer, *bufpt, *bufmax, *s, *t;
     int encoding;
     size_t bufsiz;
-    PyObject *list, *sublist, *item, *columns = NULL;
-    PyObject *(*getitem) (PyObject *, Py_ssize_t);
-    PyObject *(*getsubitem) (PyObject *, Py_ssize_t);
-    PyObject *(*getcolumn) (PyObject *, Py_ssize_t);
-    Py_ssize_t i, j, m, n = 0;
+    PyObject *rows, *iter_row, *item, *columns = NULL;
+    Py_ssize_t i, j, m, n;
 
     if (!self->cnx) {
         PyErr_SetString(PyExc_TypeError, "Connection is not valid");
@@ -709,7 +702,7 @@ conn_inserttable(connObject *self, PyObject *args)
     }
 
     /* gets arguments */
-    if (!PyArg_ParseTuple(args, "sO|O", &table, &list, &columns)) {
+    if (!PyArg_ParseTuple(args, "sO|O", &table, &rows, &columns)) {
         PyErr_SetString(
             PyExc_TypeError,
             "Method inserttable() expects a string and a list as arguments");
@@ -717,63 +710,68 @@ conn_inserttable(connObject *self, PyObject *args)
     }
 
     /* checks list type */
-    if (PyList_Check(list)) {
-        m = PyList_Size(list);
-        getitem = PyList_GetItem;
-    }
-    else if (PyTuple_Check(list)) {
-        m = PyTuple_Size(list);
-        getitem = PyTuple_GetItem;
-    }
-    else {
+    if (!(iter_row = PyObject_GetIter(rows)))
+    {
         PyErr_SetString(
             PyExc_TypeError,
-            "Method inserttable() expects a list or a tuple"
+            "Method inserttable() expects an iterable"
             " as second argument");
         return NULL;
     }
+    m = PySequence_Check(rows) ? PySequence_Size(rows) : -1;
+    if (!m) {
+        /* no rows specified, nothing to do */
+        Py_DECREF(iter_row); Py_INCREF(Py_None); return Py_None;
+    }
 
     /* checks columns type */
     if (columns) {
-        if (PyList_Check(columns)) {
-            n = PyList_Size(columns);
-            getcolumn = PyList_GetItem;
-        }
-        else if (PyTuple_Check(columns)) {
-            n = PyTuple_Size(columns);
-            getcolumn = PyTuple_GetItem;
-        }
-        else {
+        if (!(PyTuple_Check(columns) || PyList_Check(columns))) {
             PyErr_SetString(
                 PyExc_TypeError,
-                "Method inserttable() expects a list or a tuple"
+                "Method inserttable() expects a tuple or a list"
                 " as third argument");
             return NULL;
         }
+
+        n = PySequence_Fast_GET_SIZE(columns);
         if (!n) {
             /* no columns specified, nothing to do */
-            Py_INCREF(Py_None);
-            return Py_None;
+            Py_DECREF(iter_row); Py_INCREF(Py_None); return Py_None;
         }
+    } else {
+        n = -1; /* number of columns not yet known */
     }
 
     /* allocate buffer */
-    if (!(buffer = PyMem_Malloc(MAX_BUFFER_SIZE)))
-        return PyErr_NoMemory();
+    if (!(buffer = PyMem_Malloc(MAX_BUFFER_SIZE))) {
+        Py_DECREF(iter_row); return PyErr_NoMemory();
+    }
 
     encoding = PQclientEncoding(self->cnx);
 
     /* starts query */
     bufpt = buffer;
-    table = PQescapeIdentifier(self->cnx, table, strlen(table));
-    bufpt += sprintf(bufpt, "copy %s", table);
-    PQfreemem(table);
+    bufmax = bufpt + MAX_BUFFER_SIZE;
+    bufpt += snprintf(bufpt, (size_t) (bufmax - bufpt), "copy ");
+
+    s = table;
+    do {
+        t = strchr(s, '.'); if (!t) t = s + strlen(s);
+        table = PQescapeIdentifier(self->cnx, s, (size_t) (t - s));
+        if (bufpt < bufmax)
+            bufpt += snprintf(bufpt, (size_t) (bufmax - bufpt), "%s", table);
+        PQfreemem(table);
+        s = t; if (*s &&  bufpt < bufmax) *bufpt++ = *s++;
+    } while (*s);
+
     if (columns) {
         /* adds a string like f" ({','.join(columns)})" */
-        bufpt += sprintf(bufpt, " (");
+        if (bufpt < bufmax)
+            bufpt += snprintf(bufpt, (size_t) (bufmax - bufpt), " (");
         for (j = 0; j < n; ++j) {
-            PyObject *obj = getcolumn(columns, j);
-            ssize_t slen;
+            PyObject *obj = PySequence_Fast_GET_ITEM(columns, j);
+            Py_ssize_t slen;
             char *col;
 
             if (PyBytes_Check(obj)) {
@@ -781,27 +779,39 @@ conn_inserttable(connObject *self, PyObject *args)
             }
             else if (PyUnicode_Check(obj)) {
                 obj = get_encoded_string(obj, encoding);
-                if (!obj) return NULL; /* pass the UnicodeEncodeError */
+                if (!obj) {
+                    PyMem_Free(buffer); Py_DECREF(iter_row);
+                    return NULL; /* pass the UnicodeEncodeError */
+                }
                 PyBytes_AsStringAndSize(obj, &col, &slen);
                 Py_DECREF(obj);
             } else {
                 PyErr_SetString(
                     PyExc_TypeError,
                     "The third argument must contain only strings");
+                PyMem_Free(buffer); Py_DECREF(iter_row);
+                return NULL;
             }
             col = PQescapeIdentifier(self->cnx, col, (size_t) slen);
-            bufpt += sprintf(bufpt, "%s%s", col, j == n - 1 ? ")" : ",");
+            if (bufpt < bufmax)
+                bufpt += snprintf(bufpt, (size_t) (bufmax - bufpt),
+                    "%s%s", col, j == n - 1 ? ")" : ",");
             PQfreemem(col);
         }
     }
-    sprintf(bufpt, " from stdin");
+    if (bufpt < bufmax)
+        snprintf(bufpt, (size_t) (bufmax - bufpt), " from stdin");
+    if (bufpt >= bufmax)  {
+        PyMem_Free(buffer); Py_DECREF(iter_row);
+        return PyErr_NoMemory();
+    }
 
     Py_BEGIN_ALLOW_THREADS
     result = PQexec(self->cnx, buffer);
     Py_END_ALLOW_THREADS
 
-    if (!result) {
-        PyMem_Free(buffer);
+    if (!result || PQresultStatus(result) != PGRES_COPY_IN) {
+        PyMem_Free(buffer); Py_DECREF(iter_row);
         PyErr_SetString(PyExc_ValueError, PQerrorMessage(self->cnx));
         return NULL;
     }
@@ -809,33 +819,29 @@ conn_inserttable(connObject *self, PyObject *args)
     PQclear(result);
 
     /* feed table */
-    for (i = 0; i < m; ++i) {
-        sublist = getitem(list, i);
-        if (PyTuple_Check(sublist)) {
-            j = PyTuple_Size(sublist);
-            getsubitem = PyTuple_GetItem;
-        }
-        else if (PyList_Check(sublist)) {
-            j = PyList_Size(sublist);
-            getsubitem = PyList_GetItem;
-        }
-        else {
+    for (i = 0; m < 0 || i < m; ++i) {
+
+        if (!(columns = PyIter_Next(iter_row))) break;
+
+        if (!(PyTuple_Check(columns) || PyList_Check(columns))) {
+            PQendcopy(self->cnx); PyMem_Free(buffer);
+            Py_DECREF(columns); Py_DECREF(columns); Py_DECREF(iter_row);
             PyErr_SetString(
                 PyExc_TypeError,
-                "The second argument must contain a tuple or a list");
+                "The second argument must contain tuples or lists");
             return NULL;
         }
-        if (i) {
-            if (j != n) {
-                PyMem_Free(buffer);
-                PyErr_SetString(
-                    PyExc_TypeError,
-                    "Arrays contained in second arg must have same size");
-                return NULL;
-            }
-        }
-        else {
-            n = j; /* never used before this assignment */
+
+        j = PySequence_Fast_GET_SIZE(columns);
+        if (n < 0) {
+            n = j;
+        } else if (j != n) {
+            PQendcopy(self->cnx); PyMem_Free(buffer);
+            Py_DECREF(columns); Py_DECREF(iter_row);
+            PyErr_SetString(
+                PyExc_TypeError,
+                "The second arg must contain sequences of the same size");
+            return NULL;
         }
 
         /* builds insert line */
@@ -847,7 +853,7 @@ conn_inserttable(connObject *self, PyObject *args)
                 *bufpt++ = '\t'; --bufsiz;
             }
 
-            item = getsubitem(sublist, j);
+            item = PySequence_Fast_GET_ITEM(columns, j);
 
             /* convert item to string and append to buffer */
             if (item == Py_None) {
@@ -872,7 +878,8 @@ conn_inserttable(connObject *self, PyObject *args)
             else if (PyUnicode_Check(item)) {
                 PyObject *s = get_encoded_string(item, encoding);
                 if (!s) {
-                    PyMem_Free(buffer);
+                    PQendcopy(self->cnx); PyMem_Free(buffer);
+                    Py_DECREF(item); Py_DECREF(columns); Py_DECREF(iter_row);
                     return NULL; /* pass the UnicodeEncodeError */
                 }
                 else {
@@ -912,27 +919,35 @@ conn_inserttable(connObject *self, PyObject *args)
             }
 
             if (bufsiz <= 0) {
-                PyMem_Free(buffer); return PyErr_NoMemory();
+                PQendcopy(self->cnx); PyMem_Free(buffer);
+                Py_DECREF(columns); Py_DECREF(iter_row);
+                return PyErr_NoMemory();
             }
 
         }
 
+         Py_DECREF(columns);
+
         *bufpt++ = '\n'; *bufpt = '\0';
 
         /* sends data */
         if (PQputline(self->cnx, buffer)) {
             PyErr_SetString(PyExc_IOError, PQerrorMessage(self->cnx));
-            PQendcopy(self->cnx);
-            PyMem_Free(buffer);
+            PQendcopy(self->cnx); PyMem_Free(buffer);  Py_DECREF(iter_row);
             return NULL;
         }
     }
 
+    Py_DECREF(iter_row);
+    if (PyErr_Occurred()) {
+        PQendcopy(self->cnx); PyMem_Free(buffer);
+        return NULL; /* pass the iteration error */
+    }
+
     /* ends query */
     if (PQputline(self->cnx, "\\.\n")) {
         PyErr_SetString(PyExc_IOError, PQerrorMessage(self->cnx));
-        PQendcopy(self->cnx);
-        PyMem_Free(buffer);
+        PQendcopy(self->cnx); PyMem_Free(buffer);
         return NULL;
     }
 
@@ -1426,7 +1441,7 @@ static char conn_poll__doc__[] =
 "poll() -- Completes an asynchronous connection";
 
 static PyObject *
-conn_poll(connObject *self, PyObject *args)
+conn_poll(connObject *self, PyObject *noargs)
 {
     int rc;
 
@@ -1435,13 +1450,6 @@ conn_poll(connObject *self, PyObject *args)
         return NULL;
     }
 
-    /* check args */
-    if (!PyArg_ParseTuple(args, "")) {
-        PyErr_SetString(PyExc_TypeError,
-            "method poll() takes no parameters");
-        return NULL;
-    }
-
     Py_BEGIN_ALLOW_THREADS
     rc = PQconnectPoll(self->cnx);
     Py_END_ALLOW_THREADS
@@ -1612,7 +1620,7 @@ static struct PyMethodDef conn_methods[] = {
     {"describe_prepared", (PyCFunction) conn_describe_prepared,
         METH_VARARGS, conn_describe_prepared__doc__},
     {"poll", (PyCFunction) conn_poll,
-        METH_VARARGS, conn_poll__doc__},
+        METH_NOARGS, conn_poll__doc__},
     {"reset", (PyCFunction) conn_reset,
         METH_NOARGS, conn_reset__doc__},
     {"cancel", (PyCFunction) conn_cancel,
@@ -1659,7 +1667,7 @@ static struct PyMethodDef conn_methods[] = {
     {"endcopy", (PyCFunction) conn_endcopy,
         METH_NOARGS, conn_endcopy__doc__},
     {"set_non_blocking", (PyCFunction) conn_set_non_blocking,
-        METH_O, conn_set_non_blocking__doc__},
+        METH_VARARGS, conn_set_non_blocking__doc__},
     {"is_non_blocking", (PyCFunction) conn_is_non_blocking,
         METH_NOARGS, conn_is_non_blocking__doc__},
 #endif /* DIRECT_ACCESS */
